/*  
    VPPP virtual PPP connection over TCP.

    Copyright (C) 1998,1999  Maxim Krasnyansky <max_mk@yahoo.com>

    Idea is based on Sam Lantinga's PPPTCP package, but code was 
    completely written from the scratch.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
*/

/*
 * Version: 2.1 03/06/1999 Maxim Krasnyansky <max_mk@yahoo.com>
 */ 

#include <stdio.h>
#include <strings.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <signal.h>
#include <fcntl.h>
#include <syslog.h>
#include <sys/time.h>

#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <resolv.h>
#include <netdb.h>

#include "vppp.h"

int port=VPPP_PORT;
int conn_timeout=VPPP_CONN_TIMEOUT;

extern char *optarg;
extern int optind,opterr,optopt;

extern int h_errno;

char *proc_name="vppp";

struct vppp_serv{
   char host[9];
   char pass[9];
   char pppd_args[PPPD_ARGS_LEN];
} serv;

char *servs_file=VPPP_SERVS_FILE;

int read_serv(char *svr);
int auth_serv(int fd);

int main(int argc,char *argv[],char *env[]){
     int s;
     int persist=0,opt;	
     struct sockaddr_in my_addr,svr_addr;
     char *svr=NULL;
     unsigned int speed=0;
     struct hostent *hent;
     struct sigaction sa;

     while( (opt=getopt(argc,argv,"pP:t:f:")) != EOF ){
	switch(opt){
	    case 'p':
		persist=1;
		break;
	    case 'P':
	        port=atoi(optarg);	
	        break;
	    case 't':
	        conn_timeout=atoi(optarg);	
	        break;
	    case 'f':
		servs_file=strdup(optarg);
	        break;
	    default:
	        printf("Usage:vppp [-p] [-P port] [-t timeout] [-f servers_file ] <server>\n");
	        exit(1);
	}
     }	

     if( !argv[optind] ){
	printf("Usage:vppp [-p] [-P port] [-t timeout] [-f servers_file ] <server>\n");
	exit(1);
     }   	

     /* Save server name since we will modify argv array */
     svr=strdup(argv[optind]);

     openlog("vppp",LOG_PID,LOG_DAEMON);

     if( !read_serv(svr) ){	
	  syslog(LOG_ERR,"Can't find server in %s",servs_file);
	  fprintf(stderr,"Can't find server in %s\n",servs_file);
	  exit(1);
     }

     if( !(hent=gethostbyname(svr)) ){
	  syslog(LOG_ERR,"%s",hstrerror(h_errno));
	  fprintf(stderr,"%s\n",hstrerror(h_errno));
	  exit(1);
     }

     /* Become daemon */
     if(fork())
     	exit(0);
     
     close(0);close(1);close(2);
     setsid(); 
   
     memset(&sa,0,sizeof(sa));     
     sa.sa_handler=SIG_IGN;
     sigaction(SIGHUP,&sa,NULL);
     sigaction(SIGINT,&sa,NULL);
     sigaction(SIGQUIT,&sa,NULL);
     sigaction(SIGCHLD,&sa,NULL);
     sigaction(SIGPIPE,&sa,NULL);
     init_title(argc,argv,env);

     memset(&my_addr,0,sizeof(my_addr));
     my_addr.sin_family = AF_INET;
     my_addr.sin_addr.s_addr = INADDR_ANY;
     my_addr.sin_port = 0;	

     memset(&svr_addr,0,sizeof(my_addr));
     svr_addr.sin_family = AF_INET;
     svr_addr.sin_addr.s_addr = *((unsigned long *)hent->h_addr);
     svr_addr.sin_port = htons(port);	

     while(1){ 
	/* We have to create socket again each time
	 * we want to connect, since STREAM sockets 
	 * can be successfuly connected only once.
	 */
	
        if( (s=socket(AF_INET,SOCK_STREAM,0))==-1 ){
	  syslog(LOG_ERR,"Can not create socket");
	  exit(1);
        }

        if( bind(s,(struct sockaddr *)&my_addr,sizeof(my_addr)) ){
	  syslog(LOG_ERR,"Can not bind to the socket");
	  exit(1);
        }

	set_title("%s: connecting to %s",proc_name,svr);

        if( connect_t(s,(struct sockaddr *) &svr_addr, conn_timeout) ){
	  syslog(LOG_INFO,"Can't connect to the server %s",svr);
        } else {
	  if( (speed = auth_serv(s)) != -2 ){   
	     syslog(LOG_INFO,"Session %s speed %d",svr,speed);

             start_ppp(s,svr,speed,serv.pppd_args);

	     syslog(LOG_INFO,"Session %s closed",svr);
	  } else {
	     syslog(LOG_INFO,"Connection dennied by server",svr);
	  }
	}

	close(s);
	
	/* If persist option is set, sleep 
	 * and try to connect again */
	if(!persist) 
	  break;
        else
	  sleep(5);
     }
     syslog(LOG_INFO,"Exit");
     return 0;
}
	
int read_serv(char *svr)
{
     FILE *f;
     char buf[255],*str,*ptr;
     int line=0,valid=0,found=0;

     if((f=fopen(servs_file,"r"))==NULL)
        return -1;
  
     while( fgets(buf,sizeof(buf),f) && !found) {
	valid=0;
	line++;

	str=buf;
	while(*str==' ') str++;

     	if(*str== '#' || *str=='\n'){
	   valid=1; 
	   continue;
	}

	/* remove trailing garbage */
	strtok(str,"\r\n");

     	if( (str=strtok(str," \t")) == NULL)
	   break;

	/* server */

	if( strcmp(str,svr) )
	   continue;

	found=1;

     	if( (str=strtok(NULL," \t")) == NULL)
	   break;

	/* serv:pass */
	
	if( (ptr=strchr(str,':')) == NULL)
	   break;
	*ptr++='\0';

	if( strlen(str) > sizeof(serv.host)-1 )
	   break;
        strcpy(serv.host,str);

	if( strlen(ptr) > sizeof(serv.pass)-1 )
	   break;
	strcpy(serv.pass,ptr);

	/* pppd_args (might be NULL)*/

 	if( (str=strtok(NULL,"\t")) != NULL){
  	   while(*str==' ') str++;
	   if( strlen(str) > sizeof(serv.pppd_args)-1 )
	      break;
           strcpy(serv.pppd_args,str);
	}

        valid=1;
     }
     if(!valid){
        syslog(LOG_ERR,"Incorrect options on the line %d",line);
        fclose(f);
        return 0; 
     }
     return found;	
 }

#define ST_INIT  0
#define ST_HOST  1
#define ST_PASS  2

int auth_serv(int fd){
	int stage,speed=-2;
	FILE *fl;
	char buf[100];
	
	fl=fdopen(dup(fd),"w");
	setvbuf(fl,NULL,_IONBF,0);

	stage = ST_INIT;

	while( read_t(fd,buf,sizeof(buf),VPPP_TIMEOUT) > 0){
	   switch(stage){
		case ST_INIT:
	    	   if( !strncmp(buf,"VPPP",4) ){
		     stage = ST_HOST;
		     fprintf(fl,"HOST: %s\n",serv.host);
		     continue;
	           }
		   break;	

	        case ST_HOST:
		   if( !strncmp(buf,"OK",2) ){
		     stage = ST_PASS;
		     fprintf(fl,"PASS: %s\n",serv.pass);
		     continue;
	   	   }
		   break;	
	
	        case ST_PASS:
		   if(!strncmp(buf,"OK",2))
		     speed=atoi(&buf[3]);
		   break;
	   }
	
	   break;
	}

	fclose(fl);

	return speed;
}
